@@ -1,9 +1,15 @@
 #!/usr/bin/env python3
-import numpy as np
+from __future__ import annotations
 
+from abc import ABC
+from typing import List, Dict, Tuple
+
+import numpy as np
+from numpy import ndarray
 from causallearn.graph.Edge import Edge
 from causallearn.graph.Endpoint import Endpoint
 from causallearn.graph.Graph import Graph
+from causallearn.graph.Node import Node
 from causallearn.utils.GraphUtils import GraphUtils
 
 
@@ -15,28 +21,28 @@ from causallearn.utils.GraphUtils import GraphUtils
 # <->
 # --o
 # o-o
-class GeneralGraph(Graph):
+class GeneralGraph(Graph, ABC):
 
-    def __init__(self, nodes):
-        self.nodes = nodes
-        self.num_vars = len(nodes)
+    def __init__(self, nodes: List[Node]):
+        self.nodes: List[Node] = nodes
+        self.num_vars: int = len(nodes)
 
-        node_map = {}
+        node_map: Dict[Node, int] = {}
 
         for i in range(self.num_vars):
             node = nodes[i]
             node_map[node] = i
 
-        self.node_map = node_map
+        self.node_map: Dict[Node, int] = node_map
 
-        self.graph = np.zeros((self.num_vars, self.num_vars), np.dtype(int))
-        self.dpath = np.zeros((self.num_vars, self.num_vars), np.dtype(int))
+        self.graph: ndarray = np.zeros((self.num_vars, self.num_vars), np.dtype(int))
+        self.dpath: ndarray = np.zeros((self.num_vars, self.num_vars), np.dtype(int))
 
         self.reconstitute_dpath([])
 
-        self.ambiguous_triples = []
-        self.underline_triples = []
-        self.dotted_underline_triples = []
+        self.ambiguous_triples: List[Tuple[Node, Node, Node]] = []
+        self.underline_triples: List[Tuple[Node, Node, Node]] = []
+        self.dotted_underline_triples: List[Tuple[Node, Node, Node]] = []
 
         self.attributes = {}
         self.pattern = False
@@ -44,7 +50,7 @@ class GeneralGraph(Graph):
 
     ### Helper Functions ###
 
-    def adjust_dpath(self, i, j):
+    def adjust_dpath(self, i: int, j: int):
         dpath = self.dpath
         dpath[j, i] = 1
 
@@ -57,8 +63,7 @@ class GeneralGraph(Graph):
 
         self.dpath = dpath
 
-    def reconstitute_dpath(self, edges):
-
+    def reconstitute_dpath(self, edges: List[Edge]):
         for i in range(self.num_vars):
             self.adjust_dpath(i, i)
 
@@ -70,7 +75,7 @@ class GeneralGraph(Graph):
             j = self.node_map[node2]
             self.adjust_dpath(i, j)
 
-    def collect_ancestors(self, node, ancestors):
+    def collect_ancestors(self, node: Node, ancestors: List[Node]):
         if node in ancestors:
             return
 
@@ -84,7 +89,7 @@ class GeneralGraph(Graph):
     ### Public Functions ###
 
     # Adds a directed edge --> to the graph.
-    def add_directed_edge(self, node1, node2):
+    def add_directed_edge(self, node1: Node, node2: Node):
         i = self.node_map[node1]
         j = self.node_map[node2]
         self.graph[j, i] = 1
@@ -94,8 +99,7 @@ class GeneralGraph(Graph):
 
     # Adds the specified edge to the graph, provided it is not already in the
     # graph.
-    def add_edge(self, edge):
-
+    def add_edge(self, edge: Edge):
         node1 = edge.get_node1()
         node2 = edge.get_node2()
         endpoint1 = str(edge.get_endpoint1())
@@ -184,8 +188,7 @@ class GeneralGraph(Graph):
 
     # Adds a node to the graph. Precondition: The proposed name of the node
     # cannot already be used by any other node in the same graph.
-    def add_node(self, node):
-
+    def add_node(self, node: Node) -> bool:
         if node in self.nodes:
             return False
 
@@ -208,6 +211,8 @@ class GeneralGraph(Graph):
         self.graph = graph
         self.dpath = dpath
 
+        self.adjust_dpath(self.num_vars - 1, self.num_vars - 1)
+
         return True
 
     # Removes all nodes (and therefore all edges) from the graph.
@@ -221,8 +226,7 @@ class GeneralGraph(Graph):
     # Determines whether this graph contains the given edge.
     #
     # Returns true iff the graph contain 'edge'.
-    def contains_edge(self, edge):
-
+    def contains_edge(self, edge: Edge) -> bool:
         endpoint1 = str(edge.get_endpoint1())
         endpoint2 = str(edge.get_endpoint2())
 
@@ -289,12 +293,12 @@ class GeneralGraph(Graph):
     # Determines whether this graph contains the given node.
     #
     # Returns true iff the graph contains 'node'.
-    def contains_node(self, node):
+    def contains_node(self, node: Node) -> bool:
         node_list = self.nodes
         return node in node_list
 
     # Returns true iff there is a directed cycle in the graph.
-    def exists_directed_cycle(self):
+    def exists_directed_cycle(self) -> bool:
         utils = GraphUtils()
         for node in self.nodes:
             if utils.exists_directed_path_from_to_breadth_first(node, node, self):
@@ -306,8 +310,7 @@ class GeneralGraph(Graph):
     # exists if there is a directed path between the two nodes or else, for
     # some third node in the graph, there is a path to each of the two nodes in
     # question.
-    def exists_trek(self, node1, node2):
-
+    def exists_trek(self, node1: Node, node2: Node) -> bool:
         for node in self.nodes:
             if self.is_ancestor_of(node, node1) and self.is_ancestor_of(node, node2):
                 return True
@@ -322,7 +325,6 @@ class GeneralGraph(Graph):
     # an undirected edge, then in the second graph there must also be two
     # directed edges and one undirected edge between nodes A and B.
     def __eq__(self, other):
-
         if isinstance(other, GeneralGraph):
             sorted_list = self.nodes.sort()
             if sorted_list == other.nodes.sort() and np.array_equal(self.graph, other.graph):
@@ -333,10 +335,9 @@ class GeneralGraph(Graph):
             return False
 
     # Returns a mutable list of nodes adjacent to the given node.
-    def get_adjacent_nodes(self, node):
-
+    def get_adjacent_nodes(self, node: Node) -> List[Node]:
         j = self.node_map[node]
-        adj_list = []
+        adj_list: List[Node] = []
 
         for i in range(self.num_vars):
             if (not self.graph[j, i] == 0) and (not self.graph[i, j] == 0):
@@ -346,13 +347,11 @@ class GeneralGraph(Graph):
         return adj_list
 
     # Return the list of parents of a node.
-    def get_parents(self, node):
-
+    def get_parents(self, node) -> List[Node]:
         j = self.node_map[node]
-        parents = []
+        parents: List[Node] = []
 
         for i in range(self.num_vars):
-
             if (self.graph[i, j] == -1 and self.graph[j, i] == 1) \
                     or (self.graph[i, j] == Endpoint.TAIL_AND_ARROW.value
                         and self.graph[j, i] == Endpoint.ARROW_AND_ARROW.value):
@@ -362,14 +361,11 @@ class GeneralGraph(Graph):
         return parents
 
     # Returns a mutable list of ancestors for the given nodes.
-    def get_ancestors(self, nodes):
-
-        if isinstance(nodes, list):
-            pass
-        else:
+    def get_ancestors(self, nodes: List[Node]) -> List[Node]:
+        if not isinstance(nodes, list):
             raise TypeError("Must be a list of nodes")
 
-        ancestors = []
+        ancestors: List[Node] = []
 
         for node in nodes:
             self.collect_ancestors(node, ancestors)
@@ -377,10 +373,9 @@ class GeneralGraph(Graph):
         return ancestors
 
     # Returns a mutable list of children for a node.
-    def get_children(self, node):
-
+    def get_children(self, node: Node) -> List[Node]:
         i = self.node_map[node]
-        children = []
+        children: List[Node] = []
 
         for j in range(self.num_vars):
             if (self.graph[j, i] == 1 and self.graph[i, j] == -1) \
@@ -392,10 +387,8 @@ class GeneralGraph(Graph):
         return children
 
     # Returns the number of arrow endpoints adjacent to the node.
-    def get_indegree(self, node):
-
+    def get_indegree(self, node: Node) -> int:
         i = self.node_map[node]
-
         indegree = 0
 
         for j in range(self.num_vars):
@@ -408,10 +401,8 @@ class GeneralGraph(Graph):
         return indegree
 
     # Returns the number of null endpoints adjacent to the node.
-    def get_outdegree(self, node):
-
+    def get_outdegree(self, node: Node) -> int:
         i = self.node_map[node]
-
         outdegree = 0
 
         for j in range(self.num_vars):
@@ -421,14 +412,11 @@ class GeneralGraph(Graph):
         return outdegree
 
     # Returns the total number of edges into and out of the node.
-    def get_degree(self, node):
-
+    def get_degree(self, node: Node) -> int:
         i = self.node_map[node]
-
         degree = 0
 
         for j in range(self.num_vars):
-
             if self.graph[i, j] == 1 or self.graph[i, j] == -1 or self.graph[i, j] == 2:
                 degree = degree + 1
             else:
@@ -437,22 +425,21 @@ class GeneralGraph(Graph):
 
         return degree
 
-    def get_max_degree(self):
+    def get_max_degree(self) -> int:
         nodes = self.nodes
-        max = -1
+        max_degree = -1
 
         for node in nodes:
             deg = self.get_degree(node)
-            if deg > max:
-                max = deg
+            if deg > max_degree:
+                max_degree = deg
 
-        return max
+        return max_degree
 
     # Returns the node with the given string name.  In case of accidental
     # duplicates, the first node encountered with the given name is returned.
     # In case no node exists with the given name, None is returned.
-    def get_node(self, name):
-
+    def get_node(self, name: str) -> Node | None:
         for node in self.nodes:
             if node.get_name() == name:
                 return node
@@ -460,24 +447,19 @@ class GeneralGraph(Graph):
         return None
 
     # Returns the list of nodes for the graph.
-    def get_nodes(self):
+    def get_nodes(self) -> List[Node]:
         return self.nodes
 
     # Returns the names of the nodes, in the order of get_nodes.
-    def get_node_names(self):
-
-        node_names = []
-
+    def get_node_names(self) -> List[str]:
+        node_names: List[str] = []
         for node in self.nodes:
-            node_names.append(node.get_name)
-
+            node_names.append(node.get_name())
         return node_names
 
     # Returns the number of edges in the entire graph.
-    def get_num_edges(self):
-
+    def get_num_edges(self) -> int:
         edges = 0
-
         for i in range(self.num_vars):
             for j in range(i + 1, self.num_vars):
                 if self.graph[i, j] == 1 or self.graph[i, j] == -1 or self.graph[i, j] == 2:
@@ -489,14 +471,10 @@ class GeneralGraph(Graph):
         return edges
 
     # Returns the number of edges in the graph which are connected to a particular node.
-    def get_num_connected_edges(self, node):
-
+    def get_num_connected_edges(self, node: Node) -> int:
         i = self.node_map[node]
-
         edges = 0
-
         for j in range(self.num_vars):
-
             if self.graph[j, i] == 1 or self.graph[j, i] == -1 or self.graph[j, i] == 2:
                 edges = edges + 1
             else:
@@ -506,60 +484,49 @@ class GeneralGraph(Graph):
         return edges
 
     # Return the number of nodes in the graph.
-    def get_num_nodes(self):
+    def get_num_nodes(self) -> int:
         return self.num_vars
 
     # Return true iff node1 is adjacent to node2 in the graph.
-    def is_adjacent_to(self, node1, node2):
-
+    def is_adjacent_to(self, node1: Node, node2: Node) -> bool:
         i = self.node_map[node1]
         j = self.node_map[node2]
-
         return self.graph[j, i] != 0
 
     # Return true iff node1 is an ancestor of node2.
-    def is_ancestor_of(self, node1, node2):
-
+    def is_ancestor_of(self, node1: Node, node2: Node) -> bool:
         i = self.node_map[node1]
         j = self.node_map[node2]
-
         return self.dpath[j, i] == 1
 
     # Return true iff node1 is a child of node2.
-    def is_child_of(self, node1, node2):
-
+    def is_child_of(self, node1: Node, node2: Node) -> bool:
         i = self.node_map[node1]
         j = self.node_map[node2]
-
         return self.graph[i, j] == 1 or self.graph[i, j] == Endpoint.ARROW_AND_ARROW.value
 
     # Returns true iff node1 is a parent of node2.
-    def is_parent_of(self, node1, node2):
-
+    def is_parent_of(self, node1: Node, node2: Node) -> bool:
         i = self.node_map[node1]
         j = self.node_map[node2]
-
         return self.graph[j, i] == 1 or self.graph[j, i] == Endpoint.ARROW_AND_ARROW.value
 
     # Returns true iff node1 is a proper ancestor of node2.
-    def is_proper_ancestor_of(self, node1, node2):
-        return (self.is_ancestor_of(node1, node2) and not (node1 == node2))
+    def is_proper_ancestor_of(self, node1: Node, node2: Node) -> bool:
+        return self.is_ancestor_of(node1, node2) and not (node1 == node2)
 
     # Returns true iff node1 is a proper descendant of node2.
-    def is_proper_descendant_of(self, node1, node2):
-        return (self.is_descendant_of(node1, node2) and not (node1 == node2))
+    def is_proper_descendant_of(self, node1: Node, node2: Node) -> bool:
+        return self.is_descendant_of(node1, node2) and not (node1 == node2)
 
     # Returns true iff node1 is a descendant of node2.
-    def is_descendant_of(self, node1, node2):
-
+    def is_descendant_of(self, node1: Node, node2: Node) -> bool:
         i = self.node_map[node1]
         j = self.node_map[node2]
-
         return self.dpath[i, j] == 1
 
     # Returns the edge connecting node1 and node2, provided a unique such edge exists.
-    def get_edge(self, node1, node2):
-
+    def get_edge(self, node1: Node, node2: Node) -> Edge | None:
         i = self.node_map[node1]
         j = self.node_map[node2]
 
@@ -573,8 +540,7 @@ class GeneralGraph(Graph):
         return edge
 
     # Returns the directed edge from node1 to node2, if there is one.
-    def get_directed_edge(self, node1, node2):
-
+    def get_directed_edge(self, node1: Node, node2: Node) -> Edge | None:
         i = self.node_map[node1]
         j = self.node_map[node2]
 
@@ -587,18 +553,16 @@ class GeneralGraph(Graph):
         edge = Edge(node1, node2, Endpoint(end_1), Endpoint(end_2))
         return edge
 
-    # Returns the list of edges connected to a particular node. No particular ordering of the edges in the list is guaranteed.
-    def get_node_edges(self, node):
-
+    # Returns the list of edges connected to a particular node.
+    # No particular ordering of the edges in the list is guaranteed.
+    def get_node_edges(self, node: Node) -> List[Edge]:
         i = self.node_map[node]
-        edges = []
+        edges: List[Edge] = []
 
         for j in range(self.num_vars):
-
             node2 = self.nodes[j]
             if self.graph[j, i] == 1 or self.graph[j, i] == -1 or self.graph[j, i] == 2:
                 edges.append(self.get_edge(node, node2))
-
             else:
                 if self.graph[j, i] == Endpoint.TAIL_AND_ARROW.value \
                         and self.graph[i, j] == Endpoint.ARROW_AND_ARROW.value:
@@ -617,17 +581,14 @@ class GeneralGraph(Graph):
 
         return edges
 
-    def get_graph_edges(self):
-
-        edges = []
-
+    def get_graph_edges(self) -> List[Edge]:
+        edges: List[Edge] = []
         for i in range(self.num_vars):
             node = self.nodes[i]
             for j in range(i + 1, self.num_vars):
                 node2 = self.nodes[j]
                 if self.graph[j, i] == 1 or self.graph[j, i] == -1 or self.graph[j, i] == 2:
                     edges.append(self.get_edge(node, node2))
-
                 else:
                     if self.graph[j, i] == Endpoint.TAIL_AND_ARROW.value \
                             and self.graph[i, j] == Endpoint.ARROW_AND_ARROW.value:
@@ -647,7 +608,7 @@ class GeneralGraph(Graph):
         return edges
 
     # Returns the endpoint along the edge from node1 to node2, at the node2 end.
-    def get_endpoint(self, node1, node2):
+    def get_endpoint(self, node1: Node, node2: Node) -> Endpoint | None:
         edge = self.get_edge(node1, node2)
         if edge:
             return edge.get_proximal_endpoint(node2)
@@ -655,14 +616,12 @@ class GeneralGraph(Graph):
             return None
 
     # Returns true if node2 is a definite noncollider between node1 and node3.
-    def is_def_noncollider(self, node1, node2, node3):
-
+    def is_def_noncollider(self, node1: Node, node2: Node, node3: Node) -> bool:
         edges = self.get_node_edges(node2)
         circle12 = False
         circle23 = False
 
         for edge in edges:
-
             _node1 = edge.get_distal_node(node2) == node1
             _node3 = edge.get_distal_node(node2) == node3
 
@@ -681,8 +640,7 @@ class GeneralGraph(Graph):
         return False
 
     # Returns true if node2 is a definite collider between node1 and node3.
-    def is_def_collider(self, node1, node2, node3):
-
+    def is_def_collider(self, node1: Node, node2: Node, node3: Node) -> bool:
         edge1 = self.get_edge(node1, node2)
         edge2 = self.get_edge(node2, node3)
 
@@ -691,65 +649,60 @@ class GeneralGraph(Graph):
 
         return str(edge1.get_proximal_endpoint(node2)) == "ARROW" and str(edge2.get_proximal_endpoint(node2)) == "ARROW"
 
-    def is_def_unshielded_collider(self, node1, node2, node3):
-
-        return self.is_def_collider(self, node1, node2, node3) and not self.is_directly_connected_to(node1, node3)
+    def is_def_unshielded_collider(self, node1: Node, node2: Node, node3: Node) -> bool:
+        return self.is_def_collider(node1, node2, node3) and not self.is_directly_connected_to(node1, node3)
 
     # Returns true if node1 and node2 are d-connected on the set of nodes z.
-    def is_dconnected_to(self, node1, node2, z):
+    def is_dconnected_to(self, node1: Node, node2: Node, z: List[Node]) -> bool:
         utils = GraphUtils()
         return utils.is_dconnected_to(node1, node2, z, self)
 
     # Returns true if node1 and node2 are d-separated on the set of nodes z.
-    def is_dseparated_from(self, node1, node2, z):
+    def is_dseparated_from(self, node1: Node, node2: Node, z: List[Node]) -> bool:
         return not self.is_dconnected_to(node1, node2, z)
 
     # Returns true if the graph is a pattern.
-    def is_pattern(self):
+    def is_pattern(self) -> bool:
         return self.pattern
 
-    def set_pattern(self, pat):
+    def set_pattern(self, pat: bool):
         self.pattern = pat
 
     # Returns true if the graph is a PAG.
-    def is_pag(self):
+    def is_pag(self) -> bool:
         return self.pag
 
-    def set_pag(self, pag):
+    def set_pag(self, pag: bool):
         self.pag = pag
 
     # Returns true iff there is a single directed edge from node1 to node2.
-    def is_directed_from_to(self, node1, node2):
-
+    def is_directed_from_to(self, node1: Node, node2: Node) -> bool:
         i = self.node_map[node1]
         j = self.node_map[node2]
 
         return self.graph[j, i] == 1 and self.graph[i, j] == -1
 
-    # REturns true iff there is a single undirected edge between node1 and node2.
-    def is_undirected_from_to(self, node1, node2):
-
+    # Returns true iff there is a single undirected edge between node1 and node2.
+    def is_undirected_from_to(self, node1: Node, node2: Node) -> bool:
         i = self.node_map[node1]
         j = self.node_map[node2]
 
         return self.graph[j, i] == -1 and self.graph[i, j] == -1
 
-    def is_directly_connected_to(self, node1, node2):
-
+    def is_directly_connected_to(self, node1: Node, node2: Node) -> bool:
         i = self.node_map[node1]
         j = self.node_map[node2]
 
         return not (self.graph[j, i] == 0 and self.graph[i, j] == 0)
 
     # Returns true iff the given node is exogenous.
-    def is_exogenous(self, node):
+    def is_exogenous(self, node: Node) -> bool:
         return self.get_indegree(node) == 0
 
     # Returns the nodes adjacent to the given node with the given proximal endpoint.
-    def get_nodes_into(self, node, endpoint):
-
+    def get_nodes_into(self, node: Node, endpoint: Endpoint) -> List[Node]:
         i = self.node_map[node]
-        nodes = []
+        nodes: List[Node] = []
 
         if str(endpoint) == "ARROW":
             for j in range(self.num_vars):
@@ -772,10 +725,9 @@ class GeneralGraph(Graph):
         return nodes
 
     # Returns the nodes adjacent to the given node with the given distal endpoint.
-    def get_nodes_out_of(self, node, endpoint):
-
+    def get_nodes_out_of(self, node: Node, endpoint: Endpoint) -> List[Node]:
         i = self.node_map[node]
-        nodes = []
+        nodes: List[Node] = []
 
         if str(endpoint) == "ARROW":
             for j in range(self.num_vars):
@@ -798,8 +750,7 @@ class GeneralGraph(Graph):
         return nodes
 
     # Removes the given edge from the graph.
-    def remove_edge(self, edge):
-
+    def remove_edge(self, edge: Edge):
         node1 = edge.get_node1()
         node2 = edge.get_node2()
 
@@ -835,7 +786,7 @@ class GeneralGraph(Graph):
                         self.graph[j, i] = -1
                         self.graph[i, j] = 1
                     else:
-                        if end1 == 1:
+                        if end1 == -1:
                             self.graph[j, i] = Endpoint.ARROW.value
                             self.graph[i, j] = Endpoint.ARROW.value
                 else:
@@ -843,15 +794,13 @@ class GeneralGraph(Graph):
                         self.graph[j, i] = 0
                         self.graph[i, j] = 0
 
-    # Removes the edge connecting the given two nodes, provided there is exactly one such edge.
-    def remove_connecting_edge(self, node1, node2):
+        self.reconstitute_dpath(self.get_graph_edges())
 
+    # Removes the edge connecting the given two nodes, provided there is exactly one such edge.
+    def remove_connecting_edge(self, node1: Node, node2: Node):
         i = self.node_map[node1]
         j = self.node_map[node2]
 
-        out_of = self.graph[j, i]
-        in_to = self.graph[i, j]
-
         self.graph[j, i] = 0
         self.graph[i, j] = 0
 
@@ -859,8 +808,7 @@ class GeneralGraph(Graph):
     # remove at most one edge, but since multiple edges are permitted in some
     # graph implementations, the number will in some cases be greater than
     # one.
-    def remove_connecting_edges(self, node1, node2):
-
+    def remove_connecting_edges(self, node1: Node, node2: Node):
         i = self.node_map[node1]
         j = self.node_map[node2]
 
@@ -870,20 +818,18 @@ class GeneralGraph(Graph):
     # Iterates through the list and removes any permissible edges found.  The
     # order in which edges are removed is the order in which they are presented
     # in the iterator.
-    def remove_edges(self, edges):
-
+    def remove_edges(self, edges: List[Edge]):
         for edge in edges:
             self.remove_edge(edge)
 
     # Removes a node from the graph.
-    def remove_node(self, node):
-
+    def remove_node(self, node: Node):
         i = self.node_map[node]
 
         graph = self.graph
 
-        graph = np.delete(graph, (i), axis=0)
-        graph = np.delete(graph, (i), axis=1)
+        graph = np.delete(graph, i, axis=0)
+        graph = np.delete(graph, i, axis=1)
 
         self.graph = graph
 
@@ -891,33 +837,38 @@ class GeneralGraph(Graph):
         nodes.remove(node)
         self.nodes = nodes
 
-        node_map = self.node_map
-        node_map.pop(node)
+        # Node indices in node_map should be updated, so rebuild the node_map by the new nodes list
+        node_map = {}
+        for i, node in enumerate(self.nodes):
+            node_map[node] = i
         self.node_map = node_map
 
+        # num_vars should minus 1
+        self.num_vars -= 1
+
+        self.reconstitute_dpath(self.get_graph_edges())
+
     # Iterates through the list and removes any permissible nodes found.  The
     # order in which nodes are removed is the order in which they are presented
     # in the iterator.
-    def remove_nodes(self, nodes):
-
+    def remove_nodes(self, nodes: List[Node]):
         for node in nodes:
             self.remove_node(node)
 
     # Constructs and returns a subgraph consisting of a given subset of the
     # nodes of this graph together with the edges between them.
-    def subgraph(self, nodes):
-
+    def subgraph(self, nodes: List[Node]):
         subgraph = GeneralGraph(nodes)
 
         graph = self.graph
 
         for i in range(self.num_vars):
             if not (self.nodes[i] in nodes):
-                graph = np.delete(graph, (i), axis=0)
+                graph = np.delete(graph, i, axis=0)
 
         for i in range(self.num_vars):
             if not (self.nodes[i] in nodes):
-                graph = np.delete(graph, (i), axis=1)
+                graph = np.delete(graph, i, axis=1)
 
         subgraph.graph = graph
         subgraph.reconstitute_dpath(subgraph.get_graph_edges())
@@ -934,7 +885,6 @@ class GeneralGraph(Graph):
     # old graph, and this method is called to transfer the nodes and edges of
     # the old graph to the new graph.
     def transfer_nodes_and_edges(self, graph):
-
         for node in graph.nodes:
             self.add_node(node)
 
@@ -946,85 +896,85 @@ class GeneralGraph(Graph):
 
     # Returns the list of ambiguous triples associated with this graph. Triples <x, y, z> that no longer
     # lie along a path in the getModel graph are removed.
-    def get_ambiguous_triples(self):
+    def get_ambiguous_triples(self) -> List[Tuple[Node, Node, Node]]:
         return self.ambiguous_triples
 
     # Returns the set of underlines associated with this graph.
-    def get_underlines(self):
+    def get_underlines(self) -> List[Tuple[Node, Node, Node]]:
         return self.underline_triples
 
     # Returns the set of dotted underlines associated with this graph.
-    def get_dotted_underlines(self):
+    def get_dotted_underlines(self) -> List[Tuple[Node, Node, Node]]:
         return self.dotted_underline_triples
 
     # Returns true iff the triple <node1, node2, node3> is set as ambiguous.
-    def is_ambiguous_triple(self, node1, node2, node3):
+    def is_ambiguous_triple(self, node1: Node, node2: Node, node3: Node) -> bool:
         return (node1, node2, node3) in self.ambiguous_triples
 
     # Returns true iff the triple <node1, node2, node3> is set as underlined.
-    def is_underline_triple(self, node1, node2, node3):
+    def is_underline_triple(self, node1: Node, node2: Node, node3: Node) -> bool:
         return (node1, node2, node3) in self.underline_triples
 
     # Returns true iff the triple <node1, node2, node3> is set as dotted underlined.
-    def is_dotted_underline_triple(self, node1, node2, node3):
+    def is_dotted_underline_triple(self, node1: Node, node2: Node, node3: Node) -> bool:
         return (node1, node2, node3) in self.dotted_underline_triples
 
     # Adds the triple <node1, node2, node3> as an ambiguous triple to the graph.
-    def add_ambiguous_triple(self, node1, node2, node3):
+    def add_ambiguous_triple(self, node1: Node, node2: Node, node3: Node):
         self.ambiguous_triples.append((node1, node2, node3))
 
     # Adds the triple <node1, node2, node3> as an underlined triple to the graph.
-    def add_underline_triple(self, node1, node2, node3):
+    def add_underline_triple(self, node1: Node, node2: Node, node3: Node):
         self.underline_triples.append((node1, node2, node3))
 
     # Adds the triple <node1, node2, node3> as a dotted underlined triple to the graph.
-    def add_dotted_underline_triple(self, node1, node2, node3):
+    def add_dotted_underline_triple(self, node1: Node, node2: Node, node3: Node):
         self.dotted_underline_triples.append((node1, node2, node3))
 
     # Removes the triple <node1, node2, node3> from the set of ambiguous triples.
-    def remove_ambiguous_triple(self, node1, node2, node3):
+    def remove_ambiguous_triple(self, node1: Node, node2: Node, node3: Node):
         self.ambiguous_triples.remove((node1, node2, node3))
 
     # Removes the triple <node1, node2, node3> from the set of underlined triples.
-    def remove_underline_triple(self, node1, node2, node3):
+    def remove_underline_triple(self, node1: Node, node2: Node, node3: Node):
         self.underline_triples.remove((node1, node2, node3))
 
     # Removes the triple <node1, node2, node3> from the set of dotted underlined triples.
-    def remove_dotted_underline_triple(self, node1, node2, node3):
+    def remove_dotted_underline_triple(self, node1: Node, node2: Node, node3: Node):
         self.dotted_underline_triples.remove((node1, node2, node3))
 
     # Sets the list of ambiguous triples to the triples in the given set.
-    def set_ambiguous_triples(self, triples):
+    def set_ambiguous_triples(self, triples: List[Tuple[Node, Node, Node]]):
         self.ambiguous_triples = triples
 
     # Sets the list of underlined triples to the triples in the given set.
-    def set_underline_triples(self, triples):
+    def set_underline_triples(self, triples: List[Tuple[Node, Node, Node]]):
         self.underline_triples = triples
 
     # Sets the list of dotted underlined triples to the triples in the given set.
-    def set_dotted_underline_triples(self, triples):
+    def set_dotted_underline_triples(self, triples: List[Tuple[Node, Node, Node]]):
         self.dotted_underline_triples = triples
 
     # Returns a tier ordering for acyclic graphs.
-    def get_causal_ordering(self):
+    def get_causal_ordering(self) -> List[Node]:
         utils = GraphUtils()
         return utils.get_causal_order(self)
 
     # Returns true if the given node is parameterizable.
-    def is_parameterizable(self, node):
+    def is_parameterizable(self, node: Node) -> bool:
         return True
 
     # Returns true if this is a time lag model.
-    def is_time_lag_model(self):
+    def is_time_lag_model(self) -> bool:
         return False
 
     # Returns the nodes in the sepset of node1 and node2.
-    def get_sepset(self, node1, node2):
+    def get_sepset(self, node1: Node, node2: Node) -> List[Node]:
         utils = GraphUtils()
         return utils.get_sepset(node1, node2, self)
 
     # Sets the list of nodes for this graph.
-    def set_nodes(self, nodes):
+    def set_nodes(self, nodes: List[Node]):
         if len(nodes) != self.num_vars:
             raise ValueError("Sorry, there is a mismatch in the number of variables you are trying to set.")
 
@@ -1041,3 +991,6 @@ class GeneralGraph(Graph):
 
     def add_attribute(self, key, value):
         self.attributes[key] = value
+
+    def get_node_map(self) -> Dict[Node, int]:
+        return self.node_map
